import ipaddress
from typing import List, Type
from uuid import uuid4

from orchestrator.forms import FormPage, ReadOnlyField
from orchestrator.forms.validators import UniqueConstrainedList
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.workflow import StepList, done, init, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import validator
from pydantic_forms.validators import Label

from gso.products.product_blocks.iptrunk import IptrunkInterfaceBlock, IptrunkType, PhyPortCapacity
from gso.products.product_blocks.router import RouterVendor
from gso.products.product_types.iptrunk import Iptrunk
from gso.services import provisioning_proxy
from gso.services.netbox_client import NetboxClient
from gso.services.provisioning_proxy import pp_interaction
from gso.utils.helpers import (
    LAGMember,
    available_interfaces_choices,
    available_interfaces_choices_including_current_members,
    validate_iptrunk_unique_interface,
)


def initialize_ae_members(subscription: Iptrunk, initial_user_input: dict, side_index: int) -> Type[LAGMember]:
    """Initialize the list of AE members."""
    router = subscription.iptrunk.iptrunk_sides[side_index].iptrunk_side_node
    iptrunk_minimum_link = initial_user_input["iptrunk_minimum_links"]
    if router.router_vendor == RouterVendor.NOKIA:
        iptrunk_speed = initial_user_input["iptrunk_speed"]

        class NokiaLAGMember(LAGMember):
            interface_name: available_interfaces_choices_including_current_members(  # type: ignore[valid-type]
                router.owner_subscription_id,
                iptrunk_speed,
                subscription.iptrunk.iptrunk_sides[side_index].iptrunk_side_ae_members,
            ) if iptrunk_speed == subscription.iptrunk.iptrunk_speed else (
                available_interfaces_choices(router.owner_subscription_id, initial_user_input["iptrunk_speed"])
            )

        class NokiaAeMembers(UniqueConstrainedList[NokiaLAGMember]):
            min_items = iptrunk_minimum_link

        ae_members = NokiaAeMembers
    else:

        class JuniperAeMembers(UniqueConstrainedList[LAGMember]):
            min_items = iptrunk_minimum_link

        ae_members = JuniperAeMembers  # type: ignore[assignment]
    return ae_members  # type: ignore[return-value]


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    subscription = Iptrunk.from_subscription(subscription_id)

    class ModifyIptrunkForm(FormPage):
        tt_number: str
        geant_s_sid: str = subscription.iptrunk.geant_s_sid
        iptrunk_description: str = subscription.iptrunk.iptrunk_description
        iptrunk_type: IptrunkType = subscription.iptrunk.iptrunk_type
        warning_label: Label = (
            "Changing the PhyPortCapacity will result in the deletion of all AE members. "
            "You will need to add the new AE members in the next steps."  # type: ignore[assignment]
        )
        iptrunk_speed: PhyPortCapacity = subscription.iptrunk.iptrunk_speed
        iptrunk_minimum_links: int = subscription.iptrunk.iptrunk_minimum_links
        iptrunk_isis_metric: int = ReadOnlyField(subscription.iptrunk.iptrunk_isis_metric)
        iptrunk_ipv4_network: ipaddress.IPv4Network = ReadOnlyField(subscription.iptrunk.iptrunk_ipv4_network)
        iptrunk_ipv6_network: ipaddress.IPv6Network = ReadOnlyField(subscription.iptrunk.iptrunk_ipv6_network)

    initial_user_input = yield ModifyIptrunkForm
    ae_members_side_a = initialize_ae_members(subscription, initial_user_input.dict(), 0)

    class ModifyIptrunkSideAForm(FormPage):
        class Config:
            title = "Provide subscription details for side A of the trunk."

        side_a_node: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn)
        side_a_ae_iface: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_iface)
        side_a_ae_geant_a_sid: str = subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_geant_a_sid
        side_a_ae_members: ae_members_side_a = (  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members
            if initial_user_input.iptrunk_speed == subscription.iptrunk.iptrunk_speed
            else []
        )

        @validator("side_a_ae_members", allow_reuse=True)
        def validate_iptrunk_unique_interface_side_a(cls, side_a_ae_members: list[LAGMember]) -> list[LAGMember]:
            return validate_iptrunk_unique_interface(side_a_ae_members)

    user_input_side_a = yield ModifyIptrunkSideAForm
    ae_members_side_b = initialize_ae_members(subscription, initial_user_input.dict(), 1)

    class ModifyIptrunkSideBForm(FormPage):
        class Config:
            title = "Provide subscription details for side B of the trunk."

        side_b_node: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn)
        side_b_ae_iface: str = ReadOnlyField(subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_iface)
        side_b_ae_geant_a_sid: str = subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_geant_a_sid
        side_b_ae_members: ae_members_side_b = (  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members
            if initial_user_input.iptrunk_speed == subscription.iptrunk.iptrunk_speed
            else []
        )

        @validator("side_b_ae_members", allow_reuse=True)
        def validate_iptrunk_unique_interface_side_b(cls, side_b_ae_members: list[LAGMember]) -> list[LAGMember]:
            return validate_iptrunk_unique_interface(side_b_ae_members)

    user_input_side_b = yield ModifyIptrunkSideBForm

    return initial_user_input.dict() | user_input_side_a.dict() | user_input_side_b.dict()


@step("Update subscription")
def modify_iptrunk_subscription(
    subscription: Iptrunk,
    geant_s_sid: str,
    iptrunk_type: IptrunkType,
    iptrunk_description: str,
    iptrunk_speed: PhyPortCapacity,
    iptrunk_minimum_links: int,
    side_a_ae_geant_a_sid: str,
    side_a_ae_members: list[dict],
    side_b_ae_geant_a_sid: str,
    side_b_ae_members: list[dict],
) -> State:
    # Prepare the list of removed AE members
    previous_ae_members = {}
    removed_ae_members = {}
    for side_index in range(2):
        previous_ae_members[side_index] = [
            {"interface_name": member.interface_name, "interface_description": member.interface_description}
            for member in subscription.iptrunk.iptrunk_sides[side_index].iptrunk_side_ae_members
        ]
    for side_index in range(2):
        previous_members = previous_ae_members[side_index]
        current_members = side_a_ae_members if side_index == 0 else side_b_ae_members
        removed_ae_members[side_index] = [
            ae_member for ae_member in previous_members if ae_member not in current_members
        ]
    subscription.iptrunk.geant_s_sid = geant_s_sid
    subscription.iptrunk.iptrunk_description = iptrunk_description
    subscription.iptrunk.iptrunk_type = iptrunk_type
    subscription.iptrunk.iptrunk_speed = iptrunk_speed
    subscription.iptrunk.iptrunk_minimum_links = iptrunk_minimum_links

    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_geant_a_sid = side_a_ae_geant_a_sid
    #  Flush the old list of member interfaces
    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members.clear()
    #  And update the list to only include the new member interfaces
    for member in side_a_ae_members:
        subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members.append(
            IptrunkInterfaceBlock.new(subscription_id=uuid4(), **member)
        )

    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_geant_a_sid = side_b_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members.clear()
    for member in side_b_ae_members:
        subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members.append(
            IptrunkInterfaceBlock.new(subscription_id=uuid4(), **member)
        )

    subscription.description = f"IP trunk, geant_s_sid:{geant_s_sid}"

    return {
        "subscription": subscription,
        "removed_ae_members": removed_ae_members,
        "previous_ae_members": previous_ae_members,
    }


@step("Provision IP trunk interface [DRY RUN]")
def provision_ip_trunk_iface_dry(
    subscription: Iptrunk, process_id: UUIDstr, tt_number: str, removed_ae_members: List[str]
) -> State:
    provisioning_proxy.provision_ip_trunk(
        subscription, process_id, tt_number, "trunk_interface", True, removed_ae_members
    )

    return {
        "subscription": subscription,
        "label_text": "[DRY RUN] Provisioning trunk interface, please refresh to get the results of the playbook.",
    }


@step("Provision IP trunk interface [FOR REAL]")
def provision_ip_trunk_iface_real(
    subscription: Iptrunk, process_id: UUIDstr, tt_number: str, removed_ae_members: List[str]
) -> State:
    provisioning_proxy.provision_ip_trunk(
        subscription, process_id, tt_number, "trunk_interface", False, removed_ae_members
    )

    return {
        "subscription": subscription,
        "label_text": "Provisioning trunk interface, please refresh to get the results of the playbook.",
    }


@step("Update interfaces in Netbox. Reserving interfaces.")
def update_interfaces_in_netbox(subscription: Iptrunk, removed_ae_members: dict, previous_ae_members: dict) -> State:
    nbclient = NetboxClient()
    for side in range(0, 2):
        if subscription.iptrunk.iptrunk_sides[side].iptrunk_side_node.router_vendor == RouterVendor.NOKIA:
            lag_interface = subscription.iptrunk.iptrunk_sides[side].iptrunk_side_ae_iface
            router_name = subscription.iptrunk.iptrunk_sides[side].iptrunk_side_node.router_fqdn
            # Free removed interfaces
            for member in removed_ae_members[str(side)]:
                nbclient.free_interface(router_name, member["interface_name"])
            # Attach physical interfaces to LAG
            # Update interface description to subscription ID
            # Reserve interfaces
            for interface in subscription.iptrunk.iptrunk_sides[side].iptrunk_side_ae_members:
                if any(
                    ae_member.get("interface_name") == interface.interface_name
                    for ae_member in previous_ae_members[str(side)]
                ):
                    continue
                nbclient.attach_interface_to_lag(
                    device_name=subscription.iptrunk.iptrunk_sides[side].iptrunk_side_node.router_fqdn,
                    lag_name=lag_interface,
                    iface_name=interface.interface_name,
                    description=str(subscription.subscription_id),
                )
                nbclient.reserve_interface(
                    device_name=subscription.iptrunk.iptrunk_sides[side].iptrunk_side_node.router_fqdn,
                    iface_name=interface.interface_name,
                )
    return {
        "subscription": subscription,
    }


@step("Allocate interfaces in Netbox")
def allocate_interfaces_in_netbox(subscription: Iptrunk, previous_ae_members: dict) -> State:
    """Allocate the LAG interfaces in NetBox.

    attach the lag interfaces to the physical interfaces detach old ones from the LAG.
    """

    for side in range(0, 2):
        nbclient = NetboxClient()
        if subscription.iptrunk.iptrunk_sides[side].iptrunk_side_node.router_vendor == RouterVendor.NOKIA:
            for interface in subscription.iptrunk.iptrunk_sides[side].iptrunk_side_ae_members:
                if any(
                    ae_member.get("interface_name") == interface.interface_name
                    for ae_member in previous_ae_members[str(side)]
                ):
                    continue
                nbclient.allocate_interface(
                    device_name=subscription.iptrunk.iptrunk_sides[side].iptrunk_side_node.router_fqdn,
                    iface_name=interface.interface_name,
                )
            # detach the old interfaces from lag
            nbclient.detach_interfaces_from_lag(
                device_name=subscription.iptrunk.iptrunk_sides[side].iptrunk_side_node.router_fqdn,
                lag_name=subscription.iptrunk.iptrunk_sides[side].iptrunk_side_ae_iface,
            )

    return {"subscription": subscription}


@workflow(
    "Modify IP Trunk interface",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def modify_trunk_interface() -> StepList:
    return (
        init
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> modify_iptrunk_subscription
        >> update_interfaces_in_netbox
        >> pp_interaction(provision_ip_trunk_iface_dry, 3)
        >> pp_interaction(provision_ip_trunk_iface_real, 3)
        >> allocate_interfaces_in_netbox
        >> resync
        >> done
    )
